/*
RI matching algorithm for computing Kocay coefficients
*/

package algorithms.matching;

import data.Graph;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.set.hash.TIntHashSet;

import java.util.BitSet;
import java.util.Iterator;
import java.util.Vector;
import java.util.HashSet;

public class RISolver
{
	//Matching state machine
	private MatchingMachine mama;
	//Target graph
	private Graph targetGraph;
	//Number of matches of query graph into target graph
	private long numMatches;
	//Induced or not?
	private boolean induced;
	
	/*
	Constructor
	*/
	public RISolver(Graph targetGraph, boolean induced)
	{
		this.targetGraph=targetGraph;
		this.induced=induced;
		numMatches=0;
	}
	
	/*
	Compute the number of matches of query graph into target graph
	*/
	public void solve(Graph queryGraph)
	{
		numMatches=0;

		BitSet[] domains=computeDomains(queryGraph);
		this.mama=new MatchingMachine(queryGraph);
		Vector<Integer>[] symmCond=queryGraph.getSymmetryConditions();

		TIntHashSet[] targetOutAdjLists=targetGraph.getOutAdjList();
		TIntHashSet[] targetInAdjLists=targetGraph.getInAdjList();
		int nofTargetNodes=targetGraph.getNumNodes();
		int i, j;
		int nof_sn=mama.nof_sn;
		int[] parent_state=mama.parent_state;
		MamaParentType[] parent_type=mama.parent_type;
		int[] candidatesIT=new int[nof_sn];
		int[][] candidates = new int[nof_sn][];
		int[] solution = new int[nof_sn];
		for(i=0; i<nof_sn; i++)
			solution[i] = -1;
		boolean[] matched = new boolean[nofTargetNodes];

		for(i=0; i<nof_sn; i++)
		{
			if(parent_type[i] == MamaParentType.PARENTTYPE_NULL)
			{
				int n = mama.map_state_to_node[i];
				candidates[i]=new int[domains[n].cardinality()];
				int k = 0;
				for(j = domains[n].nextSetBit(0); j >= 0; j = domains[n].nextSetBit(j+1))
				{
					candidates[i][k]=j;
					k++;
				}
				candidatesIT[i]=-1;
			}
		}

		int psi = -1;
		int si = 0;
		int ci;
		int sip1;
		while(si != -1)
		{
			if(psi >= si)
				matched[solution[si]] = false;
			ci = -1;
			candidatesIT[si]++;
			while(candidatesIT[si]<candidates[si].length)
			{
				ci = candidates[si][candidatesIT[si]];
				solution[si] = ci;
				if(!matched[ci]
						&& domains[mama.map_state_to_node[si]].get(ci)
						&& condCheck(si,solution,symmCond)
						&& edgesCheck(si,ci,solution,matched))
					break;
				else
					ci=-1;
				candidatesIT[si]++;
			}
			if(ci == -1)
			{
				psi = si;
				si--;
			}
			else
			{
				if(si == nof_sn -1)
				{
					numMatches++;
					psi = si;
				}
				else
				{
					matched[solution[si]] = true;
					sip1 = si+1;
					if(parent_type[sip1] != MamaParentType.PARENTTYPE_NULL)
					{
						if(parent_type[sip1] == MamaParentType.PARENTTYPE_IN)
							candidates[sip1] = targetInAdjLists[solution[parent_state[sip1]]].toArray();
						else
							candidates[sip1] = targetOutAdjLists[solution[parent_state[sip1]]].toArray();
					}
					candidatesIT[si+1]=-1;
					psi = si;
					si++;
				}
			}
		}
	}

	private BitSet[] computeDomains(Graph queryGraph)
	{
		int numQueryNodes=queryGraph.getNumNodes();
		int numTargetNodes=targetGraph.getNumNodes();
		BitSet[] domains=new BitSet[numQueryNodes];
		int i, j;
		for(i=0;i<domains.length;i++)
			domains[i]=new BitSet(numTargetNodes);

		TIntHashSet[] targetOutAdjLists=targetGraph.getOutAdjList();
		TIntHashSet[] targetInAdjLists=targetGraph.getInAdjList();
		TIntHashSet[] queryOutAdjLists=queryGraph.getOutAdjList();
		TIntHashSet[] queryInAdjLists=queryGraph.getInAdjList();
		for(i=0;i<numTargetNodes;i++)
		{
			//Find compatible query nodes and update domains
			for(j=0;j<domains.length;j++)
			{
				if(queryOutAdjLists[j].size()<=targetOutAdjLists[i].size()
						&& queryInAdjLists[j].size()<=targetInAdjLists[i].size())
				{
					domains[j].set(i);
					//System.out.println(j+"-"+i);
				}
			}
		}

		int ra, qb, rb;
		boolean notfound;
		for(int qa=0; qa<numQueryNodes; qa++)
		{
			for(ra = domains[qa].nextSetBit(0); ra >= 0; ra = domains[qa].nextSetBit(ra+1))
			{
				//for each edge qa->qb  check if exists ra->rb
				TIntIterator it=queryOutAdjLists[qa].iterator();
				while(it.hasNext())
				{
					qb = it.next();
					notfound = true;
					TIntIterator it2=targetOutAdjLists[ra].iterator();
					while(it2.hasNext())
					{
						rb = it2.next();
						if(domains[qb].get(rb))
						{
							notfound = false;
							break;
						}
					}
					if(notfound)
						domains[qa].set(ra, false);
				}
			}
		}

		return domains;
	}

	/*
	Check if symmetry breaking conditions for the currently matched query node are satisfied
	@param si: id of currently matched query node
	@param solution: set of already matched couples of query-target nodes
	@param matched: array of boolean where the i-th entry is true iff the target node i has been already matched
	*/
	public boolean condCheck(int si, int[] solution, Vector<Integer>[] symmCond)
	{
		boolean condCheck=true;
		//System.out.println(map_state_to_node[si]);
		Vector<Integer> condNode=symmCond[mama.map_state_to_node[si]];
		int ii=0;
		for(ii=0;ii<condNode.size();ii++)
		{
			int targetNode=solution[mama.map_node_to_state[condNode.get(ii)]];
			if(solution[si]<targetNode)
			{
				condCheck=false;
				break;
			}
		}
		return condCheck;
	}
	
	/*
	Check if outgoing edges from currently matched nodes in the query and in the target, also matches
	@param si: id of the query node
	@param ci: id of the target node
	@param solution: set of already matched couples of query-target nodes
	@param matched: array of boolean where the i-th entry is true iff the target node i has been already matched
	*/
	public boolean edgesCheck(int si, int ci, int[] solution, boolean[] matched)
	{
		int k=0;
		TIntHashSet[] targetOutAdjLists=targetGraph.getOutAdjList();
		TIntHashSet[] targetInAdjLists=targetGraph.getInAdjList();
		for(int me=0; me<mama.edges_sizes[si]; me++)
		{
			int source = solution[mama.edges[si][me].source];
			int target = solution[mama.edges[si][me].target];
			if(!targetOutAdjLists[source].contains(target))
				return false;
		}
		if(induced)
		{
			int count = 0;
			TIntIterator it=targetOutAdjLists[ci].iterator();
			while(it.hasNext())
			{
				if(matched[it.next()])
				{
					count++;
					if(count>mama.o_edges_sizes[si])
						return false;
				}
			}
			count=0;
			it=targetInAdjLists[ci].iterator();
			while(it.hasNext())
			{
				if(matched[it.next()])
				{
					count++;
					if(count>mama.i_edges_sizes[si])
						return false;
				}

			}
			return true;
		}
		else
			return true;
	}

	public long getNumMatches()
	{
		return numMatches;
	}
}