/*
General definition of the EDD model, independently from the motif definition (injective or multiset)
*/

package algorithms.model;

import util.MathUtility;

import java.util.Vector;

public class EDDModel
{
	//Number of nodes in the input network
	protected int numNodes;
	//Is the input graph directed or undirected?
	protected boolean directed;
	//Array of gamma powers: gamma, gamma^2, gamma^3, ...
	protected double[] gammaPowers;
	//Out-degree distribution of the input network
	protected double[] degreeDistrOut;
	//In-degree distribution of the input network
	protected double[] degreeDistrIn;
	/*Moments of out-degree distribution. 
	The i-th entry is the moment of degree i of the out-degree distribution*/
	protected double[] momentsOut;
	/*Moments of in-degree distribution. 
	The i-th entry is the moment of degree i of the in-degree distribution*/
	protected double[] momentsIn;
	
	/*
	Constructor
	@param numNodes: number of nodes in the input network
	@param directed: is the input graph directed or undirected?
	@param degreeDistrOut: out-degree distribution of the input network
	@param degreeDistrIn: in-degree distribution of the input network
	*/
	public EDDModel(int numNodes, boolean directed, double[] degreeDistrOut, double[] degreeDistrIn)
	{
		this.numNodes=numNodes;
		this.directed=directed;
		this.degreeDistrOut=degreeDistrOut;
		this.degreeDistrIn=degreeDistrIn;
	}
	
	/*
	Compute moments up to a certain degree of the out- and in-degree distributions
	@param degree: the maximum degree for which moments must be computed
	*/
	public double[] getMomentsOut(){return momentsOut;}
	public double[] getMomentsIn(){return momentsIn;}
	public void setMoments(int degree)
	{
		momentsOut=new double[degree+1];
		momentsIn=new double[degree+1];
		int i=0, j=0;
		//Compute moments of degree 0
		momentsOut[0]=1.0;
		momentsIn[0]=1.0;
		//Compute moments of degree > 0
		for(i=1;i<momentsOut.length;i++)
		{
			momentsOut[i]=0.0;
			momentsIn[i]=0.0;
			for(j=0;j<degreeDistrOut.length;j++) {
				momentsOut[i] += degreeDistrOut[j] * Math.pow(j, i);
				//System.out.print(j+": "+degreeDistrOut[j]+" - ");
			}
			for(j=0;j<degreeDistrIn.length;j++)
				momentsIn[i]+=degreeDistrIn[j]*Math.pow(j,i);
			//System.out.println("Momento IN "+i+":"+momentsIn[i]);
			//System.out.println("Momento OUT "+i+":"+momentsOut[i]);
		}
		/*System.out.print("MomentsOut: ");
		for(i=0;i<momentsOut.length;i++)
        {
            System.out.print(momentsOut[i]+",");
        }
		System.out.println("");
		System.out.print("MomentsIn: ");
        for(i=0;i<momentsOut.length;i++)
        {
            System.out.print(momentsIn[i]+",");
        }
        */

	}
	
	/*Compute gamma and its powers up to a certain exponent
	@param degree: maximum exponent for which powers of gamma must be computed
	*/
	public void setGammaPowers(int degree)
	{
		gammaPowers=new double[degree+1];
		gammaPowers[0]=1;
		gammaPowers[1]=Math.min(1.0/((numNodes-1)*momentsOut[1]),1.0/(54*54));
		//System.out.println("gamma.pow 1:"+gammaPowers[1]);

		int i=0;
		for(i=2;i<gammaPowers.length;i++) {
			gammaPowers[i] = Math.pow(gammaPowers[1], i);
			//System.out.println("gamma.pow "+i+":"+gammaPowers[i]);
		}
	}
	
	/*Pre-computations of multinomial coefficients in the variance and covariance computation
	@param k: number of nodes in the motif
	*/
	public double[] getCoeffOverlap(int k)
	{
		double[] coeffOverlap=new double[k+1];
		int[] paramMulti=new int[3];
		paramMulti[0]=numNodes-2*k;
		paramMulti[1]=paramMulti[2]=k;
		coeffOverlap[0] = MathUtility.multinomCoeff(numNodes,paramMulti);
		paramMulti=new int[4];
		int s=0;
		for(s=1;s<=k;s++)
		{
			paramMulti[0]=k-s;
			paramMulti[1]=s;
			paramMulti[2]=k-s;
			paramMulti[3]=numNodes-2*k+s;
			coeffOverlap[s]=MathUtility.multinomCoeff(numNodes,paramMulti);
		}
		return coeffOverlap;
	}

	/*
	Occurrence probability of the topology of multiset motif
	@param adjMotif: adjacency matrix of the motif
	*/
	public double computeOccProb(boolean[][] adjMotif)
	{
		int numEdgesMotif=0;
		if(directed)
		{
			//Compute in- and out-degrees of motif nodes
			int[] outDegrees=new int[adjMotif.length];
			int[] inDegrees=new int[adjMotif.length];
			int i=0,j=0;
			for(i=0;i<adjMotif.length;i++)
			{
				for(j=0;j<adjMotif.length;j++)
				{
					if(adjMotif[i][j])
					{
						numEdgesMotif++;
						outDegrees[i]++;
						inDegrees[j]++;
					}
				}
			}
			//Compute occurrence probability of the topology
			double probTopology=gammaPowers[numEdgesMotif];
			for(i=0;i<adjMotif.length;i++)
				probTopology=probTopology*momentsOut[outDegrees[i]]*momentsIn[inDegrees[i]];
			return probTopology;
		}
		else
		{
			//Compute out-degrees of motif nodes
			int[] outDegrees=new int[adjMotif.length];
			int i=0,j=0;
			for(i=0;i<adjMotif.length;i++)
			{
				for(j=i+1;j<adjMotif.length;j++)
				{
					if(adjMotif[i][j])
					{
						numEdgesMotif++;
						outDegrees[i]++;
						outDegrees[j]++;
					}
				}
			}
			//Compute occurrence probability of the topology
			double probTopology=gammaPowers[numEdgesMotif];
			for(i=0;i<adjMotif.length;i++)
				probTopology=probTopology*momentsOut[outDegrees[i]];
			return probTopology;
		}
	}

	/*
	Compute the mean (or expected) number of occurrence of the multiset motif
	according to the EDD model
	@param nrPerm: set of non redundant permutations (NRPs) of the motif topology
	*/
	public double computeMean(Vector<boolean[][]> nrPerm)
	{
		boolean[][] adjMotif=nrPerm.get(0);
		int roPerm=nrPerm.size();
		//System.out.println("NRPerm: "+roPerm);
		double binomCoeff=MathUtility.binomCoeff(numNodes,adjMotif.length);
		//System.out.println("NumNodes:"+numNodes+"   BinomCoeff: "+binomCoeff);
		double probTopology=computeOccProb(adjMotif);
		//System.out.println("occProb: "+probTopology);
		double mean=binomCoeff*roPerm*probTopology;
		return mean;
	}

	/*
	Compute the variance of the number of occurrences of the multiset motif
	according to the EDD model
	@param nrPerm: set of non redundant permutations (NRPs) of the motif
	@param setColors: multiset of node labels of the motif
	@param mean: expected number of occurrence of the multiset motif
	@param coeffOverlap: array of multinomial coefficients in the formula of the variance
	*/
	public double computeVariance(Vector<boolean[][]> nrPerm, double mean, double[] coeffOverlap)
	{
		int i=0, j=0;
		boolean[][] adjMotif=nrPerm.get(0);
		int k=adjMotif.length;

		// Compute first part of second moment
		int numberNRPermutation=nrPerm.size();
		double varPerm = numberNRPermutation*computeOccProb(adjMotif);
		double moment2 = coeffOverlap[0]*varPerm*varPerm;

		//Compute second part of second moment
		double overlappingSum=0.0;
		for(int s=1;s<=k;s++)
		{
			double probSuperMotif=0.0;
			for(i=0;i<nrPerm.size();i++)
			{
				boolean[][] adjSource=nrPerm.get(i);
				for(j=0;j<nrPerm.size();j++)
				{
					//Compute occurrence probability of super-motif
					boolean[][] adjDest=nrPerm.get(j);
					boolean[][] adjSuper=MathUtility.createSuperMotif(adjSource,adjDest,s);
					probSuperMotif+=computeOccProb(adjSuper);
				}
			}
			overlappingSum+=coeffOverlap[s]*probSuperMotif;
		}
		moment2+=overlappingSum;

		//Compute variance
		double variance=moment2-(mean*mean);
		return variance;
	}

	/*
	Pre-computation of means of all topologies with k number of nodes in the case of
	induced motifs
	@param nrPerm: array of sets of non redundant permutations (NRPs) of the motif topology.
	Each entry of the array is the set of NRPs of a certain topology.
	*/
	public double[] computeSetTopoMeans(Vector<Vector<boolean[][]>> nrPerm)
	{
		double[] setTopoMeans=new double[nrPerm.size()];
		int i=0;
		for(i=0;i<nrPerm.size();i++)
			setTopoMeans[i]=computeMean(nrPerm.get(i));
		return setTopoMeans;
	}

	/*
	Computation of the mean of an induced motif M, starting from the pre-computed means of all
	topologies with the same number of nodes of M
	@param topoMeans: non-induced means of all topologies with the same number of nodes of M
	@param setColors: multiset of node labels of M
	@param kocayVals: coefficients of the inverted Kocay matrix in the row of the matrix
	corresponding to M
	*/
	public double computeMeanInduced(double[] topoMeans, int[] kocayVals)
	{
		int i=0;
		double meanInduced=0.0;
		for(i=0;i<kocayVals.length;i++)
		{
			if(kocayVals[i]!=0)
				meanInduced+=kocayVals[i]*topoMeans[i];
		}
		return meanInduced;
	}

	/*
	Pre-computation of variances of all topologies with k number of nodes in the case of
	induced motifs
	@param nrPerm: array of sets of non redundant permutations (NRPs) of the motif.
	Each entry of the array is the set of NRPs of a certain topology.
	@param k: number of nodes of the motif
	*/
	public double[][] computeSetTopoVariances(Vector<Vector<boolean[][]>> nrPerm, int k)
	{
		double[] coeffOverlap=getCoeffOverlap(k);
		double[][] setTopoVariances=new double[nrPerm.size()][k+1];
		int i=0;
		for(i=0;i<nrPerm.size();i++)
			setTopoVariances[i]=computeTopoVariance(nrPerm.get(i),coeffOverlap);
		return setTopoVariances;
	}

	/*
	Pre-computation of all the terms in the computation of the variance of a topology in the case
	of induced motifs
	@param nrPerm: set of non redundant permutations (NRPs) of the motif.
	@param coeffOverlap: array of multinomial coefficients in the formula of the variance
	*/
	public double[] computeTopoVariance(Vector<boolean[][]> nrPerm, double[] coeffOverlap)
	{
		int i=0, j=0, l=0, m=0;
		boolean[][] adjMotif=nrPerm.get(0);
		int k=adjMotif.length;
		double[] topoVariance=new double[k+1];
		int numberNRPermutation=nrPerm.size();
		double varTopo = numberNRPermutation*computeOccProb(adjMotif);
		topoVariance[0] = coeffOverlap[0]*varTopo*varTopo;

		double overlappingSum=0.0;
		for(int s=1;s<=k;s++)
		{
			//Compute the sum of mu values over all the super-motifs over R(m)xR(m)
			double probSuperMotif=0.0;
			for(i=0;i<nrPerm.size();i++)
			{
				boolean[][] adjSource=nrPerm.get(i);
				for(j=0;j<nrPerm.size();j++)
				{
					boolean[][] adjDest=nrPerm.get(j);
					boolean[][] adjSuper=MathUtility.createSuperMotif(adjSource,adjDest,s);
					probSuperMotif+=computeOccProb(adjSuper);
				}
			}
			topoVariance[s]=coeffOverlap[s]*probSuperMotif;
		}
		return  topoVariance;
	}

	/*
	Computation of the variance of an induced motif M, starting from the pre-computed variances
	of all topologies with the same number of nodes of M
	@param nrPerm: array of sets of non redundant permutations (NRPs) of the motif.
	Each entry of the array is the set of NRPs of a certain topology
	@param setColors: multiset of node labels of M
	@param kocayVals: coefficients of the inverted Kocay matrix in the row of the matrix
	corresponding to M
	@param topoVariances: array of pre-computed variances, one for each topology
	@param topoCovariances: array of pre-computed covariances, one for each topology
	@param topoMeans: array of pre-computed means, one for each topology
	*/
	public double computeVarianceInduced(Vector<Vector<boolean[][]>> nrPerm, int[] kocayVals, double[][] topoVariances, double[][] topoCovariances, double[] topoMeans)
	{
		double varianceInduced=0.0;

		//First term
		double varTerm=0.0;
		int i=0;
		int j=0,l=0;
		for(i=0;i<kocayVals.length;i++)
		{
			if(kocayVals[i]!=0)
			{
				double mean=topoMeans[i];
				double variance=computeMoment2Preproc(topoVariances[i])-mean*mean;
				varTerm+=kocayVals[i]*kocayVals[i]*variance;
			}
		}
		varianceInduced+=varTerm;
		//System.out.println("prova0");

		//Second term
		double covarTerm=0.0;
		//System.out.println(nrPerm.size()*(nrPerm.size()-1)/2);
		for(i=0;i<nrPerm.size();i++)
		{
			if(kocayVals[i]!=0)
			{
				for(j=i+1;j<nrPerm.size();j++)
				{
					if(kocayVals[j]!=0)
					{
						double mean1=topoMeans[i];
						double mean2=topoMeans[j];
						double covariance=computeCovarianceInduced(nrPerm.get(i),nrPerm.get(j),topoCovariances[l])-mean1*mean2;
						covarTerm+=kocayVals[i]*kocayVals[j]*covariance;
					}
					l++;
				}
			}
			else
			{
				for(j=i+1;j<nrPerm.size();j++)
					l++;
			}
		}
		covarTerm=covarTerm*2;
		varianceInduced+=covarTerm;

		return varianceInduced;
	}

	/*
	Pre-computation of the expectation of the squared counts of occurrences of all multiset motifs
	with k number of nodes
	@param topoVariance: set of variances of topologies
	@param probColorsMotif: probability of observing motif labels
	@param probSuperColors: array of probability of observing labels in the super-motifs with different overlap
	*/
	public double computeMoment2Preproc(double[] topoVariance)
	{
		// Compute first part of second moment
		double res = topoVariance[0];
		//Compute second part of second moment
		double coeff=0.0;
		double overlappingSum=0.0;
		for(int s=1;s<topoVariance.length;s++)
		{
			//Compute the sum of mu values over all the super-motifs over R(m)xR(m)
			double probSuperMotif=topoVariance[s];
			overlappingSum+=topoVariance[s];
		}
		//Sum the two parts of the formula
		res+=overlappingSum;
		return  res;
	}

	/*
	Pre-computation of covariances of all topologies with k number of nodes in the case of
	induced motifs
	@param nrPerm: array of sets of non redundant permutations (NRPs) of the motif.
	Each entry of the array is the set of NRPs of a certain topology.
	@param k: number of nodes of the motif
	*/
	public double[][] computeSetTopoCovariances(Vector<Vector<boolean[][]>> nrPerm, int k)
	{
		double[] coeffOverlap=getCoeffOverlap(k);
		double[][] setTopoCovariances=new double[nrPerm.size()*(nrPerm.size()-1)/2][k+1];
		int i=0,j=0,l=0;
		for(i=0;i<nrPerm.size();i++)
		{
			for(j=i+1;j<nrPerm.size();j++)
			{
				setTopoCovariances[l]=computeTopoCovariance(nrPerm.get(i), nrPerm.get(j), coeffOverlap);
				l++;
			}
		}
		return setTopoCovariances;
	}

	/*
	Pre-computation of all the terms in the computation of the covariance between two topologies in the case
	of induced motifs
	@param nrPerm1: set of non redundant permutations (NRPs) of the first topology
	@param nrPerm2: set of non redundant permutations (NRPs) of the second topology
	@param coeffOverlap: array of multinomial coefficients in the formula of the covariance
	*/
	public double[] computeTopoCovariance(Vector<boolean[][]> nrPerm1, Vector<boolean[][]> nrPerm2, double[] coeffOverlap)
	{
		int k=nrPerm1.get(0).length;
		double[] setTopoCovariance=new double[k+1];
		double muSum=0.0;
		int i=0,j=0,l=0;
		double[] probTopo1=new double[nrPerm1.size()];
		double[] probTopo2=new double[nrPerm2.size()];
		for(i=0;i<nrPerm1.size();i++)
		{
			boolean[][] adjMotif1=nrPerm1.get(i);
			probTopo1[i]=computeOccProb(adjMotif1);
		}
		for(i=0;i<nrPerm2.size();i++)
		{
			boolean[][] adjMotif2=nrPerm2.get(i);
			probTopo2[i]=computeOccProb(adjMotif2);
		}
		for(i=0;i<nrPerm1.size();i++)
		{
			for(j=0;j<nrPerm2.size();j++)
				muSum+=probTopo1[i]*probTopo2[j];
		}
		setTopoCovariance[0]=coeffOverlap[0]*muSum;
		//System.out.println("prova1");

		for(int s=1;s<=k;s++)
		{
			double probSuperTopology=0.0;
			for(i=0;i<nrPerm1.size();i++)
			{
				boolean[][] adjSource=nrPerm1.get(i);
				for(j=0;j<nrPerm2.size();j++)
				{
					boolean[][] adjDest=nrPerm2.get(j);
					boolean[][] adjSuper=MathUtility.createSuperMotif(adjSource,adjDest,s);
					probSuperTopology+=computeOccProb(adjSuper);
				}
			}
			setTopoCovariance[s]=coeffOverlap[s]*probSuperTopology;
		}

		return setTopoCovariance;
	}

	/*
	Computation of the covariance of an induced motif M, starting from the pre-computed covariances
	of all topologies with the same number of nodes of M
	@param nrPerm1: set of non redundant permutations (NRPs) of the first topology
	@param nrPerm2: set of non redundant permutations (NRPs) of the second topology
	@param probColorsMotif: probability of observing motif labels
	@param probSuperColors: array of probability of observing labels in the super-motifs with different overlap
	*/
	public double computeCovarianceInduced(Vector<boolean[][]> nrPerm1, Vector<boolean[][]> nrPerm2, double[] topoCovariance)
	{
		double covariance=0.0;
		//First part
		covariance+=topoCovariance[0];
		//System.out.println("prova1");
		//Second part
		double overlappingSum=0.0;
		int s=0;
		for(s=1;s<topoCovariance.length;s++)
			overlappingSum+=topoCovariance[s];
		covariance+=overlappingSum;
		//System.out.println("prova2");
		return covariance;
	}
}
	