/*
Define all the methods to estimate motif frequencies and compute their p-values
*/

package algorithms.model;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TCustomHashMap;
import algorithms.matching.RISolver;
import data.Graph;
import gnu.trove.set.hash.TIntHashSet;
import util.MatrixArrayStrategy;
import util.MathUtility;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

public class MotifManager
{
    //Number of nodes of all the considered motifs
	private int motifSize;
	//Are the motifs directed or not?
    private boolean directed;
	//Are the motifs induced or non-induced?
    private boolean induced;
	
	/*
	Constructor
	@param motifSize: number of nodes of all the considered motifs
	@param directed: are the motifs directed or not?
	@param induced: are the motifs induced or non-induced?
	*/
    public MotifManager(int motifSize, boolean directed, boolean induced)
    {
        this.motifSize=motifSize;
        this.directed=directed;
        this.induced=induced;
    }

	/*
	Initialize the EDD model for the analytical computation of mean and variance
	@param net: the target graph
	*/
    public EDDModel setupEDDModel(Graph net)
    {
        int i=0;
        int numNodes=net.getNumNodes();
        double[][] distr=net.getDegreeDistributions();
        EDDModel edd=new EDDModel(numNodes,directed,distr[0],distr[1]);
        edd.setMoments(2*motifSize-2);
        edd.setGammaPowers((2*motifSize-2)*(2*motifSize-2));
        return edd;
    }

	/*
	Compute the frequency of motifs satisfying node label and topology constraints in a target graph
	@param gt: GLabTrie data structure used to compute frequencies
	@param net: target graph
	@param setTopologies: set of admitted topologies for motifs
	@param kocayMat: kocay matrix of coefficients for induced motifs
	@param mapTopoPositions: a map where each admitted topology is associated to a unique identifier
	*/
    public long[] censusGraph(Graph net, Vector<Graph> setTopologies, int minFreq)
    {
        long[] freqsMotifs=new long[setTopologies.size()];
        int i=0;
        /*RISolver ri=new RISolver(net,induced);
        for(i=0;i<setTopologies.size();i++)
        {
            Graph query=setTopologies.get(i);
            ri.solve(query);
            long numMatches=ri.getNumMatches();
            if(numMatches>=minFreq)
                freqsMotifs[i]=numMatches;
        }

         */
        return freqsMotifs;
    }

	/*
	Compute analytical p-values of motifs and store the p-values and other statistics in the final array of maps
	@param edd: EDD model used to compute analytical p-values
	@param net: target graph
	@param mapMotifsInfo: map of motif frequencies
	@param setTopologies: set of admitted topologies for motifs
	@param mapTopoPositions: a map where each admitted topology is associated to a unique identifier
	@param kocayMat: kocay matrix of coefficients for induced motifs
	@param numMotifs: number of motifs for which analytical p-values should be computed
	@param pvalTresh: minimum analytical p-value for motifs to be included in the final set of results
	*/
    public TCustomHashMap<boolean[][],double[]> computeAnalyticalPValues(EDDModel edd, long[] freqsMotifs, Vector<Graph> setTopologies, TCustomHashMap<boolean[][],Integer> mapTopoPositions, int[][] kocayMat, double pvalThresh)
    {
        TCustomHashMap<boolean[][],double[]> finalMapMotifs=new TCustomHashMap<>(new MatrixArrayStrategy());
        double[] coeffOverlap=edd.getCoeffOverlap(motifSize);
        int i=0, j=0, k=0, l=0;
        System.out.println("Computing analytical p-values... 0%");
        double status=0.1;
        double numProcessedMotifs=0.0;
        Vector<Vector<boolean[][]>> nonRedPerm=new Vector<>();
        Iterator<boolean[][]> itTopo=mapTopoPositions.keySet().iterator();
        while(itTopo.hasNext())
        {
            boolean[][] adjMotif=itTopo.next();
            nonRedPerm.add(MathUtility.getNonRedPerm(adjMotif));
        }
        double[] topoMeans=null;
        double[][] topoVariances=null;
        double[][] topoCovariances=null;
        int[][] kocayInv=null;
        if(induced)
        {
            topoMeans=edd.computeSetTopoMeans(nonRedPerm);
            //topoVariances=edd.computeSetTopoVariances(nonRedPerm,motifSize);
            //topoCovariances=edd.computeSetTopoCovariances(nonRedPerm,motifSize);
            kocayInv=MathUtility.invertMatrix(kocayMat);
        }
        for(i=0;i<freqsMotifs.length;i++)
        {
            Graph query=setTopologies.get(i);
            int numNodesQuery=query.getNumNodes();
            TIntHashSet[] outAdiacsQuery=query.getOutAdjList();
            boolean[][] adjTopo=new boolean[numNodesQuery][numNodesQuery];
            for(j=0;j<outAdiacsQuery.length;j++)
            {
                TIntIterator it=outAdiacsQuery[j].iterator();
                while(it.hasNext())
                    adjTopo[j][it.next()]=true;
            }
            int indexTopo=mapTopoPositions.get(adjTopo);
            double mean=0.0;
            double variance=0.0;
            if(induced)
            {
                mean=edd.computeMeanInduced(topoMeans,kocayInv[indexTopo]);
                //variance=edd.computeVarianceInduced(nonRedPerm,kocayInv[indexTopo],topoVariances,topoCovariances,topoMeans);
            }
            else
            {
                mean=edd.computeMean(nonRedPerm.get(indexTopo));
                //System.out.println("Mean: "+mean);
                //variance=edd.computeVariance(nonRedPerm.get(indexTopo),mean,coeffOverlap);
                //System.out.println("Variance: "+variance);
            }
            double a=(variance-mean)/(mean+variance);
            double lambda=(1-a)*mean;
            //System.out.println("a="+a);
            //System.out.println("lambda="+lambda);
            //PolyaAeppli pa=new PolyaAeppli(a,lambda);
            long occTarget=freqsMotifs[i];
            //double pval=1-pa.lowertail(occTarget);
            double pval=0;
            if(pval<-1.0E-17)
                pval=0.0;
            else if(pval>1.0)
                pval=1.0;
            if(pval<=pvalThresh)
            {
                double[] motifStats=new double[4];
                motifStats[0]=occTarget;
                motifStats[1]=mean;
                motifStats[2]=variance;
                motifStats[3]=pval;
                finalMapMotifs.put(adjTopo,motifStats);
            }
            numProcessedMotifs++;
            /*if(numProcessedMotifs/freqsMotifs.length>=status)
            {
                System.out.println("Computing analytical p-values... "+(int)(Math.round(status*100))+"%");
                status+=0.1;
            }

             */
        }
        return finalMapMotifs;
    }

	/*
	Compute coeffiecients of kocay matrix for a set of topologies with a certain number of nodes
	@param mapTopoPositions: a map where each admitted topology is associated to a unique identifier
	@param motifSize: number of nodes of the topologies
	@param directed: are the topologies directed or not?
	*/
    public int[][] buildKocayMatrix(TCustomHashMap<boolean[][],Integer> mapTopoPositions, int motifSize, boolean directed)
    {
        int[][] kocayMat=new int[mapTopoPositions.size()][mapTopoPositions.size()];
        Iterator<boolean[][]> itSource=mapTopoPositions.keySet().iterator();
        int m=0, n=0;
        while(itSource.hasNext())
        {
            boolean[][] adjQuery=itSource.next();
            Graph query=new Graph(directed,adjQuery.length);
            for(m=0;m<adjQuery.length;m++)
            {
                for(n=0;n<adjQuery.length;n++)
                {
                    if(adjQuery[m][n])
                        query.addEdge(m,n);
                }
            }
            int i=mapTopoPositions.get(adjQuery);
            Iterator<boolean[][]> itDest=mapTopoPositions.keySet().iterator();
            while(itDest.hasNext())
            {
                boolean[][] adjTarget=itDest.next();
                Graph target=new Graph(directed,adjTarget.length);
                for(m=0;m<adjTarget.length;m++)
                {
                    for(n=0;n<adjTarget.length;n++)
                    {
                        if(adjTarget[m][n])
                            target.addEdge(m,n);
                    }
                }
                int j=mapTopoPositions.get(adjTarget);
                RISolver sol=new RISolver(target,false);
                sol.solve(query);
                kocayMat[i][j]=(int)sol.getNumMatches();
                //System.out.print(kocayMat[i][j]+",");
            }
            //System.out.println();
        }
        return kocayMat;
    }

}
