/*
Graph data structure for FlashMotif
*/

package data;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.set.hash.TIntHashSet;
import util.GraphUtility;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

public class Graph
{
	
	/*Adjacency lists for graph nodes. The i-th entry of the array is the adjacency list for node with ID i. 
	The array is a set of integers, representing the IDs of adjacent nodes of node i*/
	private TIntHashSet[] outAdjList;
    private TIntHashSet[] inAdjList;
	//Is the graph directed or not?
    private boolean directed;
    private int maxDegOut;
    private int maxDegIn;

	
	/*
	Constructor
	@param directed: is the graph directed or not?
	@param numNodes: number of nodes of the graph
	*/
    public Graph(boolean directed, int numNodes)
    {
        this.directed=directed;
        this.outAdjList=new TIntHashSet[numNodes];
        this.inAdjList=new TIntHashSet[numNodes];
        for(int i=0;i<numNodes;i++)
        {
            outAdjList[i] = new TIntHashSet();
            inAdjList[i] = new TIntHashSet();
        }
        maxDegOut=0;
        maxDegOut=0;
    }
	
	/*
	Add an edge to the graph
	@param source: the first node of the edge
	@param dest: the second node of the edge
	*/
    public void addEdge(int source, int dest)
    {
        outAdjList[source].add(dest);
        inAdjList[dest].add(source);
        if(!directed)
        {
            outAdjList[dest].add(source);
            inAdjList[source].add(dest);
        }
    }
	
	/*
	Get the number of nodes in the network
	*/
	public int getMaxDegOut(){return maxDegOut;}
	public int getMaxDegIn(){ return maxDegIn;}
    public int getNumNodes()
    {
        return outAdjList.length;
    }

	/*
	Get the out-adjacency lists of nodes in the network
	*/
    public TIntHashSet[] getOutAdjList()
    {
        return outAdjList;
    }

    /*
	Get the in-adjacency lists of nodes in the network
	*/
    public TIntHashSet[] getInAdjList()
    {
        return inAdjList;
    }

	/*
	Is the graph directed or not?
	*/
    public boolean isDirected()
    {
        return directed;
    }
	
	/*
	Compute in-degree and out-degree distributions of the graph
	degreeDistr[0] is the out-degree distributions, degreeDistr[1] is the in-degree distribution
	*/
    public double[][] getDegreeDistributions()
    {
        int numNodes=outAdjList.length;
        int i=0, j=0;
		//Compute max in- and out-degrees
        int maxDegOut=0;
        int maxDegIn=0;
        for(i=0;i<outAdjList.length;i++)
        {
            if(outAdjList[i].size()>maxDegOut)
                maxDegOut=outAdjList[i].size();
            if(inAdjList[i].size()>maxDegIn)
                maxDegIn=inAdjList[i].size();
        }
		//Compute degree distributions and normalize between 0 and 1
        double[][] degreeDistr=new double[2][];
        degreeDistr[0]=new double[maxDegOut+1];
        degreeDistr[1]=new double[maxDegIn+1];
        for(i=0;i<outAdjList.length;i++)
        {
            degreeDistr[0][outAdjList[i].size()]++;
            degreeDistr[1][inAdjList[i].size()]++;
        }
        for(i=0;i<degreeDistr.length;i++)
        {
            for(j=0;j<degreeDistr[i].length;j++) {
                //System.out.print(degreeDistr[i][j]+"  -  ");
                degreeDistr[i][j] = degreeDistr[i][j] / numNodes;
            }
        }
        this.maxDegIn=maxDegIn;
        this.maxDegOut=maxDegOut;
        return degreeDistr;
    }

    /*
	Get the list of symmetry conditions for GLabTrie data structure
	@param adjMat: adjacency matrix of a graph
	*/
    public Vector<Integer>[] getSymmetryConditions()
    {
        int i, j, k;
        Vector<int[]> vv=findAutomorphisms();
        int numNodes=outAdjList.length;
        Vector<Integer>[] listCond=new Vector[numNodes];
        for(i=0;i<listCond.length;i++)
            listCond[i]=new Vector<>();
        int vvsize=vv.size();
        boolean[] broken=new boolean[vvsize];
        for(i=0;i<numNodes;i++)
        {
            for(j=0;j<vvsize;j++)
            {
                if(!broken[j] && vv.get(j)[i]!=i)
                    break;
            }
            // There are still nodes not fixed
            if(j<vvsize)
            {
                for(k=i+1;k<numNodes;k++)
                {
                    for (j=0;j<vvsize;j++)
                    {
                        if(!broken[j] && vv.get(j)[i]==k)
                        {
                            listCond[k].add(i);
                            break;
                        }
                    }
                }
            }
            for(j=0;j<vvsize;j++)
            {
                if(vv.get(j)[i]!=i)
                    broken[j]=true;
            }
        }
        return listCond;
    }

    /*
	Computes all possible automorphisms of a graph
	@param adjMat: adjacency matrix of a graph
	*/
    private Vector<int[]> findAutomorphisms()
    {
        int i, j, k;
        int g;
        int nof_nodes = outAdjList.length;
        int[] fDir = new int[nof_nodes];
        int[] fRev = new int[nof_nodes];
        for (i = 0; i < nof_nodes; i++) {
            fDir[i] = -1;
            fRev[i] = -1;
        }
        int[][] sequence = new int[nof_nodes][nof_nodes];
        for (i = 0; i < nof_nodes; i++)
        {
            for (j = 0; j < nof_nodes; j++)
            {
                if (outAdjList[i].contains(j) || outAdjList[j].contains(i))
                {
                    int numNeighs = 0;
                    for (k = 0; k < nof_nodes; k++)
                    {
                        if (outAdjList[j].contains(k))
                            numNeighs++;
                        if (outAdjList[k].contains(j) && !outAdjList[j].contains(k))
                            numNeighs++;
                    }
                    sequence[i][j] = numNeighs;
                } else {
                    sequence[i][j] = 0;
                }
            }
        }
        for (i = 0; i < nof_nodes; i++)
            Arrays.sort(sequence[i]);

        boolean[] support = new boolean[nof_nodes * nof_nodes];
        for (i = 0; i < nof_nodes * nof_nodes; i++)
            support[i] = false;
        for (i = 0; i < nof_nodes; i++)
        {
            for (j = 0; j < nof_nodes; j++)
            {
                for (k = 0; k < nof_nodes; k++)
                {
                    if (sequence[i][k] != sequence[j][k])
                        break;
                }
                if (k >= nof_nodes)
                    support[i * nof_nodes + j] = true;
            }
        }

        Vector<int[]> vv=new Vector<>();
        for (g = 0; g < nof_nodes; g++)
        {
            if (support[g * nof_nodes])
            {
                fDir[0] = g;
                fRev[g] = 0;
                int pos = 1;
                isomorphicExtensions(fDir, fRev, vv, support, pos);
                fRev[fDir[0]] = -1;
                fDir[0] = -1;
            }
        }
        return vv;
    }

    /*
	Computes all isomorphic extensions for a graph
	*/
    private void isomorphicExtensions(int[] fDir, int[] fRev, Vector<int[]> vv, boolean[] support, int pos)
    {
        int i, j;
        int nof_nodes = fDir.length;
        int[] cand = new int[nof_nodes];
        int ncand;
        int num;
        for (i = 0; i < nof_nodes; i++)
            cand[i] = -1;
        if (pos == nof_nodes)
        {
            int[] vTemp=new int[nof_nodes];
            for(i=0;i<fDir.length;i++)
                vTemp[i]=fDir[i];
            vv.add(vTemp);
        }
        else
        {
            int n, m;
            boolean flag;
            int[] count = new int[nof_nodes];
            ncand = 0;
            for (i = 0; i < nof_nodes; i++)
                count[i] = 0;
            for (i = 0; i < nof_nodes; i++)
            {
                if(fDir[i]!=-1)
                {        // find their not mapped neighbours
                    Vector<Integer> vNei=new Vector<Integer>();
                    for(j=0;j<nof_nodes;j++)
                    {
                        if(outAdjList[i].contains(j))
                            vNei.add(j);
                        if(outAdjList[j].contains(i) && !outAdjList[i].contains(j))
                            vNei.add(j);
                    }
                    num=vNei.size();
                    for(j=0;j<num;j++)
                    {
                        int neigh=vNei.get(j);
                        if(fDir[neigh]==-1)
                        {
                            if(count[neigh]==0)
                                cand[ncand++]=neigh;
                            count[neigh]++;
                        }
                    }
                }
            }
            m = 0;
            for (i = 1; i < ncand; i++)
            {
                if (count[i] > count[m])
                    m = i;
            }
            m = cand[m];
            ncand = 0;
            boolean[] already = new boolean[nof_nodes];
            for (i = 0; i < nof_nodes; i++)
                already[i] = false;
            for (i = 0; i < nof_nodes; i++)
            {
                if(fDir[i]!=-1)
                {         // find their not mapped neighbours
                    Vector<Integer> vNei=new Vector<Integer>();
                    for(j=0;j<nof_nodes;j++)
                    {
                        if(outAdjList[fDir[i]].contains(j))
                            vNei.add(j);
                        if(outAdjList[j].contains(fDir[i]) && !outAdjList[fDir[i]].contains(j))
                            vNei.add(j);
                    }
                    num=vNei.size();
                    for(j=0;j<num;j++)
                    {
                        int neigh=vNei.get(j);
                        if(!already[neigh] && fRev[neigh]==-1 && support[m*outAdjList.length+neigh])
                        {
                            cand[ncand++]=neigh;
                            already[neigh]=true;
                        }
                    }
                }
            }
            for (i=0;i<ncand;i++)
            {
                n=cand[i];
                flag=false;
                for(j=0;j<nof_nodes;j++)
                {
                    if(fDir[j]!=-1)
                    {
                        if(outAdjList[m].contains(j)!=outAdjList[n].contains(fDir[j]))
                        {
                            flag=true;
                            break;
                        }
                        else if(outAdjList[j].contains(m)!=outAdjList[fDir[j]].contains(n))
                        {
                            flag=true;
                            break;
                        }
                    }
                }
                if(!flag)
                {
                    fDir[m] = n;
                    fRev[n] = m;
                    pos++;
                    isomorphicExtensions(fDir, fRev, vv, support, pos);
                    pos--;
                    fRev[fDir[m]] = -1;
                    fDir[m] = -1;
                }
            }
        }
    }

	/*
	Print info about the graph
	*/
    public String toString()
    {
        String str="";
        int i=0,j=0;
        for(i=0;i<outAdjList.length;i++)
        {
            str+=i+" --> ";
            TIntIterator it=outAdjList[i].iterator();
            str+=it.next();
            while(it.hasNext())
                str+=","+it.next();
            str+="\n";
        }
        return str;
    }
}
