/*
Handle input and output for FlashMotif
*/

package io;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TCustomHashMap;
import data.Graph;
import gnu.trove.set.hash.TIntHashSet;
import util.GraphUtility;
import util.MatrixArrayStrategy;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.Iterator;
import java.util.Vector;

public class FileManager
{
    
	/*
	Read a graph from input file
	@param graphFile: path of the file
	*/
	public Graph readGraph(String graphFile)
    {
        Graph g=null;
        try
        {
            BufferedReader br=new BufferedReader(new FileReader(graphFile));
            String str=br.readLine();
            boolean directed=false;
            if(str.equals("directed"))
                directed=true;
            int numNodes=Integer.parseInt(br.readLine());
            g=new Graph(directed,numNodes);
            while((str=br.readLine())!=null)
            {
                String[] split=str.split("[ \t]");
                int source=Integer.parseInt(split[0]);
                int dest=Integer.parseInt(split[1]);
                g.addEdge(source,dest);
            }
            br.close();
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }
        return g;
    }
	
	/* 
	Build the set of all topologies with a certain number of nodes.
	The topologies are actually read from text files and collected in a data structure
	The method returns a map which associates the adjacency matrix of each topology to an identifier
	@param motifSize: number of nodes of the topologies that must be read
	@param directed: are the topologies directed or undirected?
	*/
    public TCustomHashMap<boolean[][],Integer> buildCompleteTopoSet(int motifSize, boolean directed)
    {
        TCustomHashMap<boolean[][],Integer> mapTopoPositions=new TCustomHashMap<>(new MatrixArrayStrategy());
        String direction="undir";
        if(directed)
            direction="dir";
        String topologyFile="out/Topologies/"+direction+motifSize+".str";
        int i=0, j=0;
        try
        {
            //BufferedReader br = new BufferedReader(new InputStreamReader(this.getClass().getResourceAsStream(topologyFile)));
            BufferedReader br = new BufferedReader(new FileReader(topologyFile));
            String str;
            while ((str = br.readLine()) != null)
            {
                boolean[][] initAdjMat = GraphUtility.getAdjMatrix(str);
				//Make the graph in canonical form
                boolean[][] finalAdjMat = GraphUtility.getCanonAdj(initAdjMat);
                /*Graph query=new Graph(directed,finalAdjMat.length);
                for(i=0;i<finalAdjMat.length;i++)
                {
                    for(j=0;j<finalAdjMat.length;j++)
                    {
                        if(finalAdjMat[i][j])
                            query.addEdge(i,j);
                    }
                }*/
                mapTopoPositions.put(finalAdjMat, 0);
            }
            br.close();
        }
        catch(Exception e)
        {
            System.out.println("File not found: "+topologyFile);
        }
        Iterator<boolean[][]> it=mapTopoPositions.keySet().iterator();
        int cont=0;
        while(it.hasNext())
        {
            mapTopoPositions.put(it.next(),cont);
            cont++;
        }
        return mapTopoPositions;
    }
	
	/*
	Build the final set of topologies
	Topologies are either read from a user-defined input file or uploaded by the previously defined map of all topologies to a specific identifier
	@param motifFile: path of user-defined motif topologies
	@param mapTopoPositions: map which associates the adjacency matrix of each topology to an identifier
	*/
    public Vector<Graph> readMotifs(String motifFile, boolean directed, TCustomHashMap<boolean[][],Integer> mapTopoPositions)
    {
        Vector<Graph> setTopologies=new Vector<>();
        int i=0, j=0;
        if(motifFile==null)
        {
            Iterator<boolean[][]> it=mapTopoPositions.keySet().iterator();
            while(it.hasNext())
            {
                boolean[][] adjMat=it.next();
                Graph query=new Graph(directed,adjMat.length);
                for(i=0;i<adjMat.length;i++)
                {
                    for(j=0;j<adjMat.length;j++)
                    {
                        if(adjMat[i][j])
                            query.addEdge(i,j);
                    }
                }
                setTopologies.add(query);
            }
        }
        else
        {
            try
            {
                BufferedReader br = new BufferedReader(new FileReader(motifFile));
                String str;
                while ((str = br.readLine()) != null)
                {
                    boolean[][] initAdjMat = GraphUtility.getAdjMatrix(str);
                    boolean[][] finalAdjMat = GraphUtility.getCanonAdj(initAdjMat);
                    Graph query=new Graph(directed,finalAdjMat.length);
                    for(i=0;i<finalAdjMat.length;i++)
                    {
                        for(j=0;j<finalAdjMat.length;j++)
                        {
                            if(finalAdjMat[i][j])
                                query.addEdge(i,j);
                        }
                    }
                    setTopologies.add(query);
                }
                br.close();
            }
            catch (Exception e) {
                System.out.println(e.getMessage());
            }
        }
        return setTopologies;
    }
	
	/*
	Write results about motifs found and relative p-values to output file
	@param mapMotifsInfo: map of final results. Array where each entry contains the map of results for a specific topology.
	Each map associates a set of motif colors for that topology to an array of statistics about the analytical model (mean, variance and p-value)
	@param setTopologies: set of motif topologies
	@param outputFile: path where to save output results
	@param extended: boolean flag used for the AnalyticalRepro code to include also statistics about the simulation-based model
	*/
    public void writeMotifResults(TCustomHashMap<boolean[][],double[]> mapMotifsInfo, String outputFile)
    {
        int numMotifs=0;
        int numTotalOccs=0;
        try
        {
            BufferedWriter bw = new BufferedWriter(new FileWriter(outputFile));
            bw.write("Topology\tNum_occ_input_graph\tMean_analytical\tVariance_analytical\tPval_analytical");
            bw.write("\n");
            Iterator<boolean[][]> it=mapMotifsInfo.keySet().iterator();
            while(it.hasNext())
            {
                boolean[][] adjTopo=it.next();
                String strTopo=GraphUtility.getAdjString(adjTopo);
                double[] motifInfo=mapMotifsInfo.get(adjTopo);
                numTotalOccs+=motifInfo[0];
                bw.write(strTopo+"\t"+(long)motifInfo[0]+"\t"+motifInfo[1]+"\t"+motifInfo[2]+"\t"+motifInfo[3]);
                bw.write("\n");
                numMotifs++;
            }
            bw.close();
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }
        System.out.println("Done! Found "+numMotifs+" motifs and a total number of "+numTotalOccs+" occurrences");
        System.out.println("Results written in "+outputFile);
    }

    /*
	Write only results about motifs found and relative number of occurrences (without p-value)
	@param mapMotifsInfo: map of final results. Array where each entry contains the map of results for a specific topology.
	Each map associates a set of motif colors for that topology to an array of statistics about the analytical model (mean, variance and p-value)
	@param setTopologies: set of motif topologies
	@param outputFile: path where to save output results
	@param extended: boolean flag used for the AnalyticalRepro code to include also statistics about the simulation-based model
	*/
    public void writeFrequenciesResults(long[] freqsMotifs, Vector<Graph> setTopologies, String outputFile)
    {
        int numTotalOccs=0;
        int numMotifs=0;
        try
        {
            BufferedWriter bw = new BufferedWriter(new FileWriter(outputFile));
            bw.write("Topology\tNum_occ_input_graph\n");
            int i=0, j=0;
            for(i=0;i<setTopologies.size();i++)
            {
                Graph query=setTopologies.get(i);
                int numNodesQuery=query.getNumNodes();
                TIntHashSet[] outAdiacsQuery=query.getOutAdjList();
                boolean[][] adjTopo=new boolean[numNodesQuery][numNodesQuery];
                for(j=0;j<outAdiacsQuery.length;j++)
                {
                    TIntIterator it=outAdiacsQuery[j].iterator();
                    while(it.hasNext())
                        adjTopo[j][it.next()]=true;
                }
                String strTopo=GraphUtility.getAdjString(adjTopo);
                numTotalOccs+=freqsMotifs[i];
                bw.write(strTopo+"\t"+freqsMotifs[i]+"\n");
                if(freqsMotifs[i]>0)
                    numMotifs++;
            }
            bw.close();
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }
        System.out.println("Done! Found "+numMotifs+" motifs and a total number of "+numTotalOccs+" occurrences");
        System.out.println("Results written in "+outputFile);
    }

}
