/*
Collection of methods for matemathical computations
*/

package util;

import java.util.Vector;

public class MathUtility
{
	//Precomputed binomial coefficients from (0 0) to (10 10). The [i,j]-th entry is the binomial coefficient (i j)
	private static int matBinomCoeff[][]={ 
	{1,  0,  0,   0,   0,   0,   0,   0,  0,  0, 0} ,
	{1,  1,  0,   0,   0,   0,   0,   0,  0,  0, 0} ,
	{1,  2,  1,   0,   0,   0,   0,   0,  0,  0, 0} , 
	{1,  3,  3,   1,   0,   0,   0,   0,  0,  0, 0} , 
	{1,  4,  6,   4,   1,   0,   0,   0,  0,  0, 0} , 
	{1,  5, 10,  10,   5,   1,   0,   0,  0,  0, 0} ,
	{1,  6, 15,  20,  15,   6,   1,   0,  0,  0, 0} ,
	{1,  7, 21,  35,  35,  21,   7,   1,  0,  0, 0} ,
	{1,  8, 28,  56,  70,  56,  28,   8,  1,  0, 0} ,
	{1,  9, 36,  84, 126, 126,  84,  36,  9,  1, 0} ,
	{1, 10, 45, 120, 210, 252, 210, 120, 45, 10, 1} 
	};
	
	/*
	Compute the binomial coefficient (n k)
	*/
	public static double binomCoeff(int n, int k)
	{
		if(k>n)
			return 0;
		else 
		{
			if(k==0)
				return 1;
			else if(k==1)
				return n; 
			else if(n<=10) {
				return matBinomCoeff[n][k];
			} 
			else 
			{
				//Vandermonde identity
				if(k>(n/2))
					k=n-k;
				int m1=n/2;
				int m2=n-m1;
				double res=0;
				if(m1!=m2) 
					// Pascal triangle
					res=binomCoeff(2*m1,k)+binomCoeff(2*m1,k-1);
				else 
				{
					int k_end ;
					if(k%2==1) 
						//Odd
						k_end=(k+1)/2;
					else 
					{
						//even
						k_end=k/2;
						res=binomCoeff(m1,k_end );
						res=res*res;
					}
					for(int i=0;i<k_end;i++)
						res+=2*(binomCoeff(m1,i)*binomCoeff(m1,k-i));
				}
				return res;
			}
		}
	}
	
	/*
	Compute the multinomial coefficient (n k[0], k[1], k[2], ..., k[x])
	*/
	public static double multinomCoeff(int n,int[] k) 
	{
		int max=-1;
		int i_max=-1;
		for(int i=0;i<k.length;i++) 
		{
			if(k[i]>max ) 
			{
				i_max=i;
				max=k[i];
			}
		}
		double res=factorial(n,max);
		double denom=1;
		for(int i=0;i<k.length;i++) 
		{
			if(i!=i_max)
				denom=denom*factorial(k[i],1); 
		}
		res=res/denom;
		return res;
	}
	
	/*
	Compute n!/k! = n*(n-1)*...*(k+2)*(k+1)
	*/
	public static double factorial(int n,int k) 
	{
		double res = (n>=(k+1))? (k + 1) : 1 ; 
		for(int i=k+2;i<=n;i++) 
			res*= i;
		return res;
	}
	
	/*
	Compute the next permutation of indexes after the current one
	@param p: the current permutation of indexes
	*/
	public static boolean nextPermutation(int[] p) 
	{
		for (int a = p.length - 2; a >= 0; --a)
		{
			if (p[a] < p[a + 1])
			{
				for (int b = p.length - 1;; --b)
				{
					if (p[b] > p[a]) 
					{
						int t = p[a];
						p[a] = p[b];
						p[b] = t;
						for (++a, b = p.length - 1; a < b; ++a, --b) 
						{
							t = p[a];
							p[a] = p[b];
							p[b] = t;
						}
						return true;
					}
				}
			}
		}
		return false;
	}
	
	/*
	Compute the adjMatrix of a super-motif as a result of overlapping of two motifs with s nodes
	@param adjSource: adjacency matrix of the first motif
	@param adjDest: adjacency matrix of the second motif
	@param s: number of overlapping nodes
	*/
	public static boolean[][] createSuperMotif(boolean[][] adjSource, boolean[][] adjDest, int s)
	{
		int k=adjSource.length;
		int l=0, m=0;
		boolean[][] adjSuper=new boolean[2*k-s][2*k-s];
		//Copy adjSource in the super-motif adjMatrix
		for(l=0;l<adjSource.length;l++)
			for(m=0;m<adjSource.length;m++)
				adjSuper[l][m]=adjSource[l][m];
		//Copy adjDest in the super-motif adjMatrix
		for(l=0;l<adjDest.length;l++)
			for(m=0;m<adjDest.length;m++)
				adjSuper[l+k-s][m+k-s]=adjDest[l][m];
		//Compute logical OR in the middle part of super-motif adjMatrix
		for(l=k-s;l<k;l++)
			for(m=k-s;m<k;m++)
				adjSuper[l][m]=adjSuper[l][m] || adjSource[l][m];
		return adjSuper;
	}

	/*
	Compute the inverse of a matrix (if it exists) through the LU decomposition method
	@param mat: input matrix
	*/
	public static int[][] invertMatrix(int[][] mat)
	{
		//Build identity matrix
		double[][] identityMat=new double[mat.length][mat.length];
		int i=0, j=0, k=0;
		for(i=0;i<identityMat.length;i++)
			identityMat[i][i]=1.0;
		
		//Make a LU decomposition of input matrix
		double[][] LU=new double[mat.length][mat.length];
		for(i=0;i<mat.length;i++)
		{
			for(j=0;j<mat.length;j++)
				LU[i][j]=mat[i][j];
		}
		int m=mat.length;
		int n=mat.length;
		int[] piv=new int[m];
		for(i=0;i<m;i++)
			piv[i]=i;
		int pivsign=1;
		double[] LUrowi;
		double[] LUcolj=new double[m];
		for(j=0;j<n;j++) 
		{
			for(i=0;i<m;i++) 
			{
				LUcolj[i]=LU[i][j];
			}
			for(i=0;i<m;i++) 
			{
				LUrowi=LU[i];
				int kmax=Math.min(i,j);
				double s=0.0;
				for(k=0;k<kmax;k++) 
				{
					s += LUrowi[k]*LUcolj[k];
				}
				LUrowi[j]=LUcolj[i]-=s;
			}
			int p=j;
			for(i=j+1;i<m;i++) 
			{
				if(Math.abs(LUcolj[i])>Math.abs(LUcolj[p])) 
				{
					p=i;
				}
			}
			if(p!=j)
			{
				for(k=0;k<n;k++)
				{
					double t=LU[p][k]; 
					LU[p][k]=LU[j][k];
					LU[j][k]=t;
				}
				k=piv[p];
				piv[p]=piv[j];
				piv[j]=k;
				pivsign=-pivsign;
			}
			if(j<m & LU[j][j]!=0.0) 
			{
				for(i=j+1;i<m;i++) 
				{
					LU[i][j]/=LU[j][j];
				}
			}
		}
		
		//Solve with respect to identity matrix and LU decomposition
		int nx=identityMat.length;
		double[][] X=new double[mat.length][mat.length];
		for(i=0;i<piv.length;i++)
		{
			for(j=0;j<nx;j++)
				X[i][j]=identityMat[piv[i]][j];
		}
		for(k=0;k<n;k++)
		{
			for(i=k+1;i<n;i++)
			{
				for(j=0;j<nx;j++) 
				{
					X[i][j]-=X[k][j]*LU[i][k];
				}
			}
		}
		for(k=n-1;k>=0;k--)
		{
			for(j=0;j<nx;j++) 
			{
				X[k][j]/=LU[k][k];
			}
			for(i=0;i<k;i++) 
			{
				for(j=0;j<nx;j++)
				{
					X[i][j]-=X[k][j]*LU[i][k];
				}
			}
		}
		
		/*for(i=0;i<X.length;i++)
		{
			for(j=0;j<X.length;j++)
				System.out.print(X[i][j]+",");
			System.out.println();
		}*/
		
		//Return the inverse matrix with integer values
		int[][] inverseMat=new int[X.length][X.length];
		for(i=0;i<inverseMat.length;i++)
		{
			for(j=0;j<inverseMat.length;j++)
				inverseMat[i][j]=(int)(Math.round(X[i][j]));
		}
		return inverseMat;
		
	}
	
	/*
	Compute the set of non-redundant permutations of a motif topology, represented through its adjacency matrix
	@param adjMotif: adjacency matrix of the motif topology
	*/
	public static Vector<boolean[][]> getNonRedPerm(boolean[][] adjMotif)
	{
		Vector<boolean[][]> nonRedPerm=new Vector<boolean[][]>();
		int[] currentPerm=new int[adjMotif.length];
		int i=0, j=0, k=0;
		for(i=0; i<currentPerm.length;i++)
			currentPerm[i]=i;
		//Add the indentity permutation to the set of non redundant permutations
		nonRedPerm.add(adjMotif);
		//Compute all possible permutations and add the corrisponding motif to the non redundant list, if not present
		int numPerm=1;
		for(;MathUtility.nextPermutation(currentPerm);) 
		{
			//System.out.println(numPerm);
			//Get the adjacency matrix corresponding to the new permutation
			boolean[][] newAdjMotif=new boolean[adjMotif.length][adjMotif.length];
			for(i=0;i<currentPerm.length;i++)
			{
				for(j=0;j<currentPerm.length;j++)
					newAdjMotif[currentPerm[i]][currentPerm[j]]=adjMotif[i][j];
			}
			//Check if the two matrices are different
			for(k=0;k<nonRedPerm.size();k++)
			{
				boolean[][] checkAdjMotif=nonRedPerm.get(k);
				boolean diff=false;
				i=0;
				while(!diff && i<checkAdjMotif.length)
				{
					j=0;
					while(!diff && j<checkAdjMotif.length)
					{
						if(newAdjMotif[i][j]!=checkAdjMotif[i][j])
							diff=true;
						j++;
					}
					i++;
				}
				if(!diff)
					break;
			}
			//Motif is new, so it can be added to the non redundant set
			if(k==nonRedPerm.size())
				nonRedPerm.add(newAdjMotif);
			numPerm++;
		}
		return nonRedPerm;
	}

}
	